// ----------------------------------------------------------------------------
/// count.c
/// © 2009-2012 Sealevel Systems, Inc.
/// <summary>
/// Sample demonstrating use of R9 Development board's digital outputs.
/// </summary>
///
/// <remarks>
/// This application acts as a simple binary counter.
/// 
/// Command line arguments are accepted allowing you to specifiy the maximum
/// value the outputs should count to as well as the delay (in ms) between each
/// count.  If no arguments are specified, the maximum attainable with the
/// number of outputs available will be used with a 1 ms pause between each count. 
/// </remarks>
///
/// <TODO>
/// The method of calculating default maximum count won't work beyond 32 outputs, 
/// so if we ever make an R9 variant with more than 32 outputs, we need to link
/// the C math lib.
/// </TODO>
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "gpio.h"

#ifdef BOARD_R91001
#include "r91001_map.h"
#elif BOARD_R92100
#include "r92100_map.h"
#elif BOARD_S91X00
#include "s91x00_map.h"
#else
#error "UNKNOWN BOARD TYPE!!!!"
#endif

int main(int argc, char *args[])
{
	int i;
	int length = DIGITAL_OUTPUT_MAP_LENGTH;
	int count = (int) ((1 << length) - 1);
	int delay = 1000;

	// We will accept any arguments as indication of the
	// count and delay
	if (argc > 1)
	{
		int reCount = atoi(args[1]);
		if (reCount > 0 && reCount < count)
		{
			count = reCount;
		}
	}
	if (argc > 2)
	{
		int reDelay = atoi(args[2]);
		if (reDelay > 0 && reDelay < 5000)
		{
			delay = 1000 * reDelay;
		}
	}

	printf("binary count to %d with %d ms pause between each count...\n", count, delay / 1000);

	// register the point(s)
	for (i = 0; i < length; i++)
	{
		if (DIGITAL_OUTPUT_MAP[i] < 0)
		{
			continue;
		}

		if (gpio_register(DIGITAL_OUTPUT_MAP[i]) != 0)
		{
			printf("Error registering output %d with sysfs!\n", i);
			return -1;
		}

		// Ensure points are configured as outputs
		if (gpio_set_direction(DIGITAL_OUTPUT_MAP[i], "out") != 0)
		{
			printf("Error setting direction of output %d!\n", i);
			return -2;
		}
	}

	// count and display the value in binary on all available outputs
	for (i = 0; i < count; i++)
	{
		int j, bit = 0;

		for (j = 0; j < length; j++)
		{
			if (DIGITAL_OUTPUT_MAP[j] < 0)
			{
				continue;
			}

			// set the value of output j to the corresponding binary bit of i
			if (gpio_set_value(DIGITAL_OUTPUT_MAP[j], ((i >> bit++) & 0x01)) != 0)
			{
				printf("Error setting value of output %d!\n", j);
				return -4;
			}
		}
		
		// now delay for some ms (delay is actually stored in us, but specified in ms)
		usleep(delay);
	}

	// if you no longer wish to access the points, you can unregister them
	for (i = 0; i < length; i++)
	{
		if (DIGITAL_OUTPUT_MAP[i] < 0)
		{
			continue;
		}

		gpio_unregister(DIGITAL_OUTPUT_MAP[i]);
	}
	
	return 0;
}
