// ----------------------------------------------------------------------------
/// updowncount.c
/// © 2009-2012 Sealevel Systems, Inc.
/// <summary>
/// Sample demonstrating use of R9 Development board's quad counters in Up-Down
/// counter mode.
/// </summary>
///
/// <remarks>
/// This application acts as a simple Up-Down counter status display.  The act
/// of reading the counters on the R9 development board is accomplished through
/// the SPI interface.  The counters on the R9 (LSI 7366) can be configured for
/// multiple methods of operation.  This sample uses the counters in the simplest
/// mode possible.  The counter may be configured to signal events through ARM
/// GPIO inputs as well as through constant polling, as is demonstrated in this
/// example.
///
/// The counters are configued for non-quadrature (fixed direction based on B), 
/// free count mode, with the A pin acting as the clock, B as the direction, and
/// the Index pin acting as a counter reset.  Be sure to wire the counter inputs
/// to appropriate values (GND, 3.3V) for your setup or they will be floating 
/// and cause unpredictable results.  Note that A and B are positive polarity
/// while I is negative polarity.
/// 
/// No command line arguments are accepted.  This tool simply displays the
/// status of all R9 counters.
/// </remarks>
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <asm/types.h>

#include <linux/spi/spidev.h>

#include "r91001_map.h"
#include "ls7366.h"

int main(int argc, char *args[])
{
	int i;

	// Query the value of each counter (and make sure it is configured right)
	printf("Counter | Up/Down | Value\n");
	for (i = 0; i < COUNTER_DEV_LENGTH; i++)
	{
		struct spi_ioc_transfer transaction[2];
		unsigned char txbuffer[2];
		unsigned char status;
		signed long count;
		int fd;
		
		// Open the SPI device corresponding to the nth counter
		fd = open(COUNTER_DEV[i], O_RDWR, 0);
		if (fd <= 0)
		{
			printf("Couldn't open counter %i!\n", i);
			return -1;
		}

		memset(transaction, 0x00, sizeof(transaction));
		transaction[0].tx_buf = (__u64) txbuffer;

		// Check current status
		txbuffer[0] = INSTRUCTION_CMD_READ | INSTRUCTION_REG_STR;
		transaction[0].len = 1;
		transaction[1].rx_buf = (__u64) &status;
		transaction[1].len = 1;
		ioctl(fd, SPI_IOC_MESSAGE(2), transaction);

		// Clear the status reg
		txbuffer[0] = INSTRUCTION_CMD_CLEAR | INSTRUCTION_REG_STR;
		write(fd, &txbuffer, 1);

		// Configure the counter, mode register 0
		txbuffer[0] = INSTRUCTION_CMD_WRITE | INSTRUCTION_REG_MDR0;
		txbuffer[1] = MDR0_INDEX_RESET_CNTR | MDR0_MODE_FREE | MDR0_NONQUAD;
		write(fd, &txbuffer, 2);

		// Configure the counter, mode register 1
		txbuffer[0] = INSTRUCTION_CMD_WRITE | INSTRUCTION_REG_MDR1;
		txbuffer[1] = MDR1_COUNTER_ENABLE | MDR1_WIDTH_4B;
		write(fd, &txbuffer, 2);

		// Finally read the current value of CNTR
		txbuffer[0] = INSTRUCTION_CMD_READ | INSTRUCTION_REG_CNTR;
		transaction[1].rx_buf = (__u64) ((unsigned char *) &count);
		transaction[0].len = 1;
		transaction[1].len = 4;

		ioctl(fd, SPI_IOC_MESSAGE(2), transaction);
		
		// The count is returned in reverse byte order, rearrange it
		count = ((count & 0x000000FF) << 24) | ((count & 0x0000FF00) << 8) | ((count & 0x00FF0000) >> 8) | ((count & 0xFF000000) >> 24);

		printf("%2i       | %1i     | %lu\n", (i + 1), (status & STR_DIRECTION) > 0, count);

		close(fd);
	}

	return 0;
}
